@@ -5,13 +5,56 @@ __author__ = "neutronics material maker development team"
 import json
 import warnings
 from pathlib import Path
+from typing import Optional
 
 try:
     import openmc
 except BaseException:
     warnings.warn(
-        "OpenMC not found, .openmc_material, .serpent_material, .mcnp_material,\
-            .fispact_material not avaiable")
+        "OpenMC not found, .openmc_material, .serpent_material,"
+        " .mcnp_material, .fispact_material .shif_materials not avaiable")
+
+
+def check_add_additional_end_lines(value):
+    """Uses to check the additional lines passed to Material and Multimaterial
+    classes are correctly formatted"""
+
+    if value is not None:
+        string_codes = ['mcnp', 'serpent', 'shift', 'fispact']
+        if not isinstance(value, dict):
+            raise ValueError(
+                'Material.additional_end_lines should be a dictionary')
+        for key, entries in value.items():
+            if key not in string_codes:
+                raise ValueError(
+                    'Material.additional_end_lines should be a '
+                    'dictionary where the keys are the name of the neutronics'
+                    'code. Acceptable values are {}'.format(string_codes))
+            if not isinstance(entries, list):
+                raise ValueError(
+                    'Material.additional_end_lines should be a'
+                    ' dictionary where the value of each dictionary entry is a'
+                    ' list')
+            for entry in entries:
+                if not isinstance(entry, str):
+                    raise ValueError(
+                        'Material.additional_end_lines should be'
+                        'a dictionary where the value of each dictionary entry'
+                        ' is a list of strings')
+    return value
+
+
+def add_additional_end_lines(code: str, mat) -> list:
+    """
+    Accertains if additional lines were requested by the user for the code used
+    and if so returns the additional lines request as a list to be added to the
+    end of the existing material card list.
+    """
+    print(mat.additional_end_lines)
+    if mat.additional_end_lines is not None:
+        if code in list(mat.additional_end_lines.keys()):
+            return mat.additional_end_lines[code]
+    return []
 
 
 def make_fispact_material(mat) -> str:
@@ -40,6 +83,8 @@ def make_fispact_material(mat) -> str:
         atoms = mat.volume_in_cm3 * atoms_cm3
         mat_card.append(isotope + " " + "{:.12E}".format(atoms))
 
+    mat_card = mat_card + add_additional_end_lines('fispact', mat)
+
     return "\n".join(mat_card)
 
 
@@ -75,6 +120,8 @@ def make_serpent_material(mat) -> str:
             + f"{isotope[1]:.{mat.decimal_places}e}"
         )
 
+    mat_card = mat_card + add_additional_end_lines('serpent', mat)
+
     return "\n".join(mat_card)
 
 
@@ -124,6 +171,8 @@ def make_mcnp_material(mat) -> str:
 
         mat_card.append(start + rest)
 
+    mat_card = mat_card + add_additional_end_lines('mcnp', mat)
+
     return "\n".join(mat_card)
 
 
@@ -145,7 +194,8 @@ def make_shift_material(mat) -> str:
         name = mat.material_tag
 
     mat_card = [
-        "name %s\n" % name
+        "[COMP][MATERIAL]\n"
+        + "name %s\n" % name
         + "matid %s\n" % mat.material_id
         + "tmp %s" % mat.temperature_in_K
     ]
@@ -158,6 +208,9 @@ def make_shift_material(mat) -> str:
         nd_ += ' ' + f"{atom_dens[1]:.{mat.decimal_places}e}"
 
     mat_card.extend([zaid, nd_])
+
+    mat_card = mat_card + add_additional_end_lines('shift', mat)
+
     return "\n".join(mat_card)
 
 
@@ -187,7 +240,7 @@ def AddMaterialFromDir(directory: str, verbose: bool = True):
             print(sorted(list(new_data.keys())), "\n")
 
 
-def AddMaterialFromFile(filename, verbose=True) -> None:
+def AddMaterialFromFile(filename: str, verbose: Optional[bool] = True) -> None:
     """Add materials to the internal library from a json file"""
     with open(filename, "r") as f:
         new_data = json.load(f)
